import homekeeper.lib
import homekeeper.testing


cd = homekeeper.lib.cd
lib = homekeeper.lib
BaseFakeFilesystemTest = homekeeper.testing.BaseFakeFilesystemTest


class LibTest(BaseFakeFilesystemTest):
    def test_change_directory(self, os):
        current_directory = os.getcwd()
        directory = os.path.join(os.sep, 'foo', 'bar')
        lib.makedirs(directory)
        with cd(directory):
            assert os.getcwd() == directory
        assert os.getcwd() == current_directory

    def test_remove_directory(self, fopen, os):
        directory = os.path.join(os.sep, 'foo', 'bar')
        lib.makedirs(directory)
        with fopen(os.path.join(directory, 'file.txt'), 'w') as f:
            f.write('hello!')
        assert os.path.exists(directory)
        lib.remove(directory)
        assert not os.path.exists(directory)

    def test_makedirs_does_not_raise(self, os):
        directory = os.path.join(os.sep, 'foo', 'bar')
        lib.makedirs(directory)
        assert os.path.exists(directory)
        lib.makedirs(directory)
        assert os.path.exists(directory)

    def test_get_home_directory(self, os):
        assert os.getenv('HOME') == os.path.join(os.sep, 'home', 'johndoe')

    def test_remove_file_or_symlink(self, os):
        source = os.path.join(self.home_directory, '.bashrc')
        target = os.path.join(self.home_directory, '.bash_profile')
        os.symlink(source, target)
        lib.remove(source)
        lib.remove(target)
        assert not os.path.exists(source)
        assert not os.path.exists(target)



import logging
import mock
import pyfakefs
import pytest
import click.testing
import homekeeper.cli
import homekeeper.config
import homekeeper.lib


config = homekeeper.config
lib = homekeeper.lib
CliRunner = click.testing.CliRunner
ConfigData = homekeeper.config.ConfigData


class BaseFakeFilesystemTest:
    """A base test class for interacting with a fake filesystem.  Working with a real filesystem is possible; however,
    two problems need to be resolved:

    1) The build will fail on Windows; a Linux subsystem for Windows must be installed as a workaround.
    2) Resolving os.getenv('HOME') will be problematic; all calls must be patched to return a tmpdir path.
    """
    def setup_method(self):
        self.fake_fs = pyfakefs.fake_filesystem.FakeFilesystem()
        self.fake_fopen = pyfakefs.fake_filesystem.FakeFileOpen(self.fake_fs)
        self.fake_os = pyfakefs.fake_filesystem.FakeOsModule(self.fake_fs)
        self.fake_shutil = pyfakefs.fake_filesystem_shutil.FakeShutilModule(self.fake_fs)
        self.home_directory = None
        self.dotfiles_directory = None
        self.directories = []
        self.files = []
        self.patchers = []
        self.setup_patchers()
        self.setup_fs(self.fake_os)

    def setup_patchers(self):
        logging.info('patching homekeeper')
        self.patchers.append(mock.patch('homekeeper.config.fopen', self.fake_fopen))
        self.patchers.append(mock.patch('homekeeper.config.os', self.fake_os))
        self.patchers.append(mock.patch('homekeeper.lib.fopen', self.fake_fopen))
        self.patchers.append(mock.patch('homekeeper.lib.os', self.fake_os))
        self.patchers.append(mock.patch('homekeeper.lib.shutil', self.fake_shutil))
        logging.info('patching homekeeper.symlink')
        self.patchers.append(mock.patch('homekeeper.symlink.os', self.fake_os))
        self.patchers.append(mock.patch('homekeeper.symlink.shutil', self.fake_shutil))
        # Note that only `shutildisk_usage()` is faked, according to fake_filesystem_shutil.py.  The rest of the
        # functions will work only if `os` and `open` are patched as well.
        logging.info('patching shutil')
        self.patchers.append(mock.patch('shutil.open', self.fake_fopen))
        self.patchers.append(mock.patch('shutil.os', self.fake_os))
        for patcher in self.patchers:
            patcher.start()

    def setup_fs(self, os):
        self.home_directory = os.path.join(os.sep, 'home', 'johndoe')
        self.dotfiles_directory = os.path.join(self.home_directory, 'dotfiles')
        self.directories = [
            self.home_directory,
            self.dotfiles_directory,
            os.path.join(self.dotfiles_directory, '.vim')
        ]
        self.files = [
            os.path.join(self.dotfiles_directory, '.bashrc'),
            os.path.join(self.dotfiles_directory, '.vimrc'),
        ]
        logging.info("setting HOME environment variable in fake os: %s", self.home_directory)
        os.environ['HOME'] = str(self.home_directory)
        logging.info("setting up directories in fake fs: %s", self.directories)
        for directory in self.directories:
            lib.makedirs(directory)
        logging.info("setting up files in fake fs: %s", self.files)
        for file in self.files:
            lib.touch(file)

    def setup_file_symlink(self, filename):
        os = self.fake_os
        source = os.path.join(self.dotfiles_directory, filename)
        target = os.path.join(self.home_directory, filename)
        lib.touch(source)
        lib.remove(target)
        os.symlink(source, target)
        return source, target

    def setup_directory_symlink(self, dirname):
        os = self.fake_os
        source = os.path.join(self.dotfiles_directory, dirname)
        target = os.path.join(self.home_directory, dirname)
        lib.makedirs(source)
        lib.remove(target)
        os.symlink(source, target)
        return source, target

    def teardown_method(self):
        for patcher in self.patchers:
            patcher.stop()
        del self.fake_fs

    @pytest.fixture
    def fopen(self):
        return self.fake_fopen

    @pytest.fixture
    def os(self):
        return self.fake_os

    @pytest.fixture
    def shutil(self):
        return self.fake_shutil

    @pytest.fixture
    def files(self):
        files = [
            '.bashrc',
            '.bash_profile',
            '.gitconfig',
            '.gitignore',
            '.vimrc',
        ]
        for filename in files:
            lib.touch(self.dotfiles_directory, filename)
        return files

    @pytest.fixture
    def directories(self):
        directories = [
            'bin',
            '.i3',
            '.vim',
        ]
        for dirname in directories:
            lib.makedirs(self.dotfiles_directory, dirname)
        return directories


class BaseCliTest(BaseFakeFilesystemTest):
    def setup_method(self):
        super().setup_method()
        os = self.fake_os
        self.runner = CliRunner()
        self.config_data = ConfigData()
        self.config_data.directories = [os.path.join(self.dotfiles_directory)]

    def run(self, *args):
        result = self.runner.invoke(homekeeper.cli.main, args)
        logging.info("stdout: %s", result.stdout)
        return result



import homekeeper.cli
import homekeeper.testing


cli = homekeeper.cli
lib = homekeeper.lib
BaseCliTest = homekeeper.testing.BaseCliTest


class CliCleanupTest(BaseCliTest):
    def test_cleanup_does_not_raise(self):
        self.run('cleanup')

    def test_cleanup_ignores_flags(self, os):
        dotfile = os.path.join(lib.get_home_directory(), '.dotfile')
        dotfile_link = os.path.join(lib.get_home_directory(), '.dotfile.link')
        lib.touch(dotfile)
        os.symlink(dotfile, dotfile_link)
        assert os.path.exists(dotfile)
        assert os.path.exists(dotfile_link)
        lib.remove(dotfile)
        result = self.run('--debug',
                          '--no-cleanup-symlinks',
                          '--no-overwrite',
                          '--config-path=/home/johndoe/.homekeeper.json',
                          'cleanup')
        assert result.exit_code == 0
        assert not os.path.exists(dotfile)
        assert not os.path.exists(dotfile_link)



import logging
import sys
import click
import homekeeper
import homekeeper.exceptions


ConfigException = homekeeper.exceptions.ConfigException
Homekeeper = homekeeper.Homekeeper
HomekeeperException = homekeeper.exceptions.HomekeeperException


@click.command(short_help='removes broken symlinks only')
@click.pass_context
def cleanup(ctx):
    try:
        h = Homekeeper(ctx.obj)
        if not h.cleanup_symlinks:
            logging.warning('read flag --no-cleanup-symlinks but forcing value to be true')
            h.cleanup_symlinks = True
        h.cleanup()
    except (ConfigException, HomekeeperException) as e:
        logging.error(e)
        sys.exit(1)


@click.command(short_help='symlinks dotfiles to your home directory')
@click.pass_context
def keep(ctx):
    try:
        h = Homekeeper(ctx.obj)
        h.keep()
    except (ConfigException, HomekeeperException) as e:
        logging.error(e)
        sys.exit(1)


@click.command(short_help='restores dotfiles and replacing symlinks')
@click.pass_context
def unkeep(ctx):
    try:
        h = Homekeeper(ctx.obj)
        h.unkeep()
    except (ConfigException, HomekeeperException) as e:
        logging.error(e)
        sys.exit(1)


@click.command(short_help='displays the version and then exits')
@click.pass_context
def version(ctx):
    try:
        h = Homekeeper(ctx.obj)
        click.echo('homekeeper version %s' % h.version())
    except (ConfigException, HomekeeperException) as e:
        logging.error(e)
        sys.exit(1)


@click.group()
@click.option('--cleanup-symlinks/--no-cleanup-symlinks',
              default=True,
              is_flag=True,
              help='removes broken symlinks (default true)')
@click.option('--overwrite/--no-overwrite',
              default=True,
              is_flag=True,
              help='overwrite existing files or symlinks (default true)')
@click.option('--config-path',
              metavar='FILE',
              default=None,
              help='path to .homekeeper.json config file')
@click.option('--debug/--no-debug',
              default=False,
              is_flag=True,
              help='enables debug output (default false)')
@click.pass_context
def main(ctx, cleanup_symlinks, overwrite, config_path, debug):
    if debug:
        logging.basicConfig(level=logging.DEBUG)
        logging.info('set log level to debug')
    else:
        logging.basicConfig(format='%(message)s', level=logging.INFO)
    ctx.obj = {
        'cleanup_symlinks': cleanup_symlinks,
        'config_path': config_path,
        'overwrite': overwrite,
    }
    logging.debug("created click context object: %s", ctx.obj)


main.add_command(cleanup)
main.add_command(keep)
main.add_command(unkeep)
main.add_command(version)



import homekeeper.config
import homekeeper.testing


config = homekeeper.config
lib = homekeeper.lib
BaseCliTest = homekeeper.testing.BaseCliTest
ConfigData = homekeeper.config.ConfigData


# pylint: disable=unused-argument
class CliKeepTest(BaseCliTest):
    def test_keep(self, os, files, directories):
        config.write(self.config_data)
        result = self.run('--debug', 'keep')
        assert result.exit_code == 0

    def test_keep_without_cleanup(self, os, files, directories):
        config.write(self.config_data)
        result = self.run('--debug', '--no-cleanup-symlinks', 'keep')
        # TODO: Add assertions for no cleanup
        assert result.exit_code == 0

    def test_keep_without_overwriting(self, os, files, directories):
        config.write(self.config_data)
        result = self.run('--debug', '--no-overwrite', 'keep')
        # TODO: Add assertions for no no overwrite
        assert result.exit_code == 0

    def test_keep_with_different_config_path(self, os, files, directories):
        lib.makedirs(self.home_directory, 'config')
        config_file = os.path.join(self.home_directory, 'config', '.homekeeper.json')
        config.write(self.config_data, config_file)
        result = self.run('--debug', "--config-path={}".format(config_file), 'keep')
        assert result.exit_code == 0

    def test_keep_with_multiple_source_directories(self, os, files, directories):
        dotfiles_1 = os.path.join(self.home_directory, 'dotfiles')
        dotfiles_2 = os.path.join(self.home_directory, 'dotfiles2')
        lib.makedirs(dotfiles_1)
        lib.makedirs(dotfiles_2)
        self.config_data.directories = [str(pathname) for pathname in [dotfiles_1, dotfiles_2]]
        config.write(self.config_data)
        result = self.run('--debug', 'keep')
        assert result.exit_code == 0



import os
import pytest
import homekeeper.config
import homekeeper.exceptions
import homekeeper.lib
import homekeeper.testing


config = homekeeper.config
fopen = homekeeper.lib.fopen
testing = homekeeper.testing
ConfigData = homekeeper.config.ConfigData
ConfigException = homekeeper.exceptions.ConfigException


class ConfigTest:
    def get_module_directory(self):
        return os.path.dirname(__file__)

    def get_data(self, *args):
        return os.path.join(self.get_module_directory(), *args)

    def test_read(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read')
        config_data = config.read(config_file)
        assert config_data.directories == [
            '/home/johndoe/dotfiles/desktop',
            '/home/johndoe/dotfiles/laptop',
        ]
        assert config_data.excludes == [
            '.git',
            'README.md'
        ]

    def test_read_directory(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read_directory')
        config_data = config.read(config_file)
        assert config_data.directories == ['/home/johndoe/dotfiles']

    def test_read_single_directory(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read_single_directory')
        config_data = config.read(config_file)
        assert config_data.directories == ['/home/johndoe/dotfiles']

    def test_read_invalid_single_directory(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read_invalid_single_directory')
        with pytest.raises(ConfigException):
            config.read(config_file)

    def test_read_invalid_directories(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read_invalid_directories')
        with pytest.raises(ConfigException):
            config.read(config_file)

    def test_read_invalid_excludes(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read_invalid_excludes')
        with pytest.raises(ConfigException):
            config.read(config_file)

    def test_read_invalid_exclude(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read_invalid_single_exclude')
        with pytest.raises(ConfigException):
            config.read(config_file)

    def test_read_invalid_file_with_bad_decoding(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read_file_with_bad_decoding')
        with pytest.raises(ConfigException):
            config.read(config_file)

    def test_read_missing_file(self):
        config_file = self.get_data('testdata', 'ConfigTest.test_read_missing_file')
        with pytest.raises(ConfigException):
            config.read(config_file)

    def test_read_with_both_directory_and_directories_fails(self):
        # TODO: Implement me!
        pass

    def test_read_directory_with_non_str_value_fails(self):
        # TODO: Implement me!
        pass

    def test_write(self, tmpdir):
        config_data = ConfigData()
        config_data.directories = ['/home/johndoe/dotfiles']
        config_data.excludes = ['.git']
        config_file = config.write(config_data, os.path.join(tmpdir.strpath, 'homekeeper.json'))
        with fopen(config_file, encoding='utf-8') as f:
            actual = f.read()
        test_file = os.path.join(self.get_module_directory(), 'testdata', 'ConfigTest.test_write')
        with fopen(test_file, encoding='utf-8') as f:
            expected = f.read()
        assert actual == expected



import logging
import homekeeper.config
import homekeeper.lib
import homekeeper.symlink


__version__ = '5.0.3'
config = homekeeper.config
lib = homekeeper.lib
symlink = homekeeper.symlink
ConfigData = homekeeper.config.ConfigData


class Homekeeper:
    def __init__(self, ctx):
        self.overwrite = ctx['overwrite']
        self.cleanup_symlinks = ctx['cleanup_symlinks']
        self.config_path = ctx['config_path']
        logging.debug("read context: %s", ctx)

    def cleanup(self):
        logging.debug('beginning cleanup')
        symlink.cleanup_symlinks(lib.get_home_directory())
        logging.debug('finished cleanup')

    def keep(self):
        logging.debug('beginning keep')
        config_data = config.read(self.config_path)
        for directory in config_data.directories:
            symlink.create_symlinks(
                directory,
                lib.get_home_directory(),
                config_data=config_data,
                overwrite=self.overwrite)
        if self.cleanup_symlinks:
            self.cleanup()
        logging.debug('finished keep')

    def unkeep(self):
        logging.debug('beginning unkeep')
        config_data = config.read(self.config_path)
        for directory in config_data.directories:
            symlink.restore_symlinks(
                directory,
                lib.get_home_directory(),
                config_data=config_data)
        if self.cleanup_symlinks:
            self.cleanup()
        logging.debug('finished unkeep')

    def version(self):
        return __version__



import homekeeper.config
import homekeeper.lib
import homekeeper.symlink
import homekeeper.testing

lib = homekeeper.lib
symlink = homekeeper.symlink
BaseFakeFilesystemTest = homekeeper.testing.BaseFakeFilesystemTest
ConfigData = homekeeper.config.ConfigData


class SymlinkRestoreTest(BaseFakeFilesystemTest):
    def test_restore_symlink_with_file(self, os):
        source, target = self.setup_file_symlink('.bashrc')
        symlink.restore_symlink(source, target)
        assert not os.path.islink(target)
        assert os.path.isfile(target)

    def test_restore_symlink_with_unlinked_target(self, os):
        source, target = self.setup_file_symlink('.bashrc')
        os.unlink(target)
        symlink.restore_symlink(source, target)
        assert os.path.exists(source)
        assert not os.path.exists(target)

    def test_restore_symlink_with_same_source_and_target(self, os):
        source, target = self.setup_file_symlink('.bashrc')
        symlink.restore_symlink(target, target)
        assert os.path.islink(target)
        assert source == os.readlink(target)

    def test_restore_symlink_with_directory(self, os):
        source, target = self.setup_directory_symlink('.vim')
        lib.touch(source, 'autoload', 'pathogen.vim')
        symlink.create_symlink(source, target, overwrite=True)
        symlink.restore_symlink(source, target)
        assert not os.path.islink(target)
        assert os.path.isdir(target)
        assert os.path.isfile(os.path.join(target, 'autoload', 'pathogen.vim'))

    def test_restore_symlinks(self, os, files, directories):
        config_data = ConfigData()
        config_data.excludes = []
        symlink.create_symlinks(self.dotfiles_directory, self.home_directory, config_data=config_data, overwrite=True)
        symlink.restore_symlinks(self.dotfiles_directory, self.home_directory, config_data=config_data)
        for pathname in files + directories:
            source = os.path.join(self.dotfiles_directory, pathname)
            target = os.path.join(self.home_directory, pathname)
            assert os.path.exists(source)
            assert os.path.exists(target)
            assert not os.path.islink(source)
            assert not os.path.islink(target)

    def test_restore_symlinks_with_deleted_targets(self, os, files, directories):
        config_data = ConfigData()
        config_data.excludes = []
        symlink.create_symlinks(self.dotfiles_directory, self.home_directory, config_data=config_data, overwrite=True)
        for pathname in files + directories:
            lib.remove(self.home_directory, pathname)
        symlink.restore_symlinks(self.dotfiles_directory, self.home_directory, config_data=config_data)
        for pathname in files + directories:
            source = os.path.join(self.dotfiles_directory, pathname)
            target = os.path.join(self.home_directory, pathname)
            assert os.path.exists(source)
            assert not os.path.exists(target)



import errno
import logging
import os
import shutil
import homekeeper.exceptions


HomekeeperException = homekeeper.exceptions.HomekeeperException


class cd:
    def __init__(self, pathname):
        self.pathname = pathname
        self.saved_pathname = None

    def __enter__(self):
        self.saved_pathname = os.getcwd()
        logging.debug("changing directory: %s", self.pathname)
        os.chdir(self.pathname)

    def __exit__(self, etype, value, traceback):
        logging.debug("changing directory: %s", self.saved_pathname)
        os.chdir(self.saved_pathname)


def fopen(*args, **kwargs):
    return open(*args, **kwargs)


def get_current_working_directory():
    return os.path.realpath(os.getcwd())


def get_home_directory():
    return os.getenv('HOME')


def makedirs(*args):
    try:
        directory = os.path.join(*args)
        logging.debug("making directory: %s", directory)
        os.makedirs(directory)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(directory):
            pass
        else:
            raise


def remove(*args):
    try:
        target = os.path.join(*args)
        if os.path.islink(target):
            os.unlink(target)
            logging.debug("removed symlink: %s", target)
            return
        if os.path.isfile(target):
            os.remove(target)
            logging.debug("removed file: %s", target)
            return
        if os.path.isdir(target):
            shutil.rmtree(target)
            logging.debug("removed directory: %s", target)
            return
    except IOError:
        logging.debug("could not remove path: %s", target)


def touch(*args):
    target = os.path.join(*args)
    makedirs(os.path.dirname(target))
    with fopen(target, 'a', encoding='utf-8'):
        logging.debug("touching path: %s", target)
        os.utime(target)


def process_directories(source_directory, target_directory, config_data=None, overwrite=True, process=None):
    if config_data is None:
        raise HomekeeperException('no value set for config_data')
    if not process:
        raise HomekeeperException('no value set for process')
    if not os.path.isdir(source_directory):
        logging.info("dotfiles directory not found: %s", source_directory)
        return
    if source_directory == target_directory:
        logging.error("source and target directory are the same: %s", source_directory)
        return
    logging.info("processing paths in directory: %s", source_directory)
    with cd(source_directory):
        logging.debug("excluding paths: %s", config_data.excludes)
        for pathname in os.listdir('.'):
            basename = os.path.basename(pathname)
            source = os.path.join(source_directory, basename)
            target = os.path.join(target_directory, basename)
            if basename in config_data.excludes:
                logging.debug("skipping because path is excluded: %s", basename)
                continue
            logging.debug("processing %s -> %s", source, target)
            process(source, target, overwrite=overwrite)



import homekeeper.testing


lib = homekeeper.lib
BaseCliTest = homekeeper.testing.BaseCliTest


class CliVersionTest(BaseCliTest):
    def test_version(self):
        result = self.run('version')
        assert result.exit_code == 0
        assert result.output == "homekeeper version {}\n".format(homekeeper.__version__)



import homekeeper.lib
import homekeeper.symlink
import homekeeper.testing


cd = homekeeper.lib.cd
lib = homekeeper.lib
symlink = homekeeper.symlink
BaseFakeFilesystemTest = homekeeper.testing.BaseFakeFilesystemTest


class SymlinkCleanupTest(BaseFakeFilesystemTest):
    def test_cleanup_symlinks(self, os):
        lib.touch(self.home_directory, '.a.txt')
        lib.touch(self.home_directory, '.b.txt')
        with cd(self.home_directory):
            os.symlink('.a.txt', '.link-a.txt')
            os.symlink('.b.txt', '.link-b-1.txt')
            os.symlink('.b.txt', '.link-b-2.txt')
            assert os.path.exists('.link-a.txt')
            assert os.path.exists('.link-b-1.txt')
            assert os.path.exists('.link-b-2.txt')
            lib.remove(self.home_directory, '.b.txt')
            symlink.cleanup_symlinks(self.home_directory)
            assert os.path.exists('.link-a.txt')
            assert not os.path.exists('.link-b-1.txt')
            assert not os.path.exists('.link-b-2.txt')

    def test_cleanup_symlinks_ignores_non_dotfiles(self):
        # TODO: Implement me!
        pass



import json
import logging
import os
import homekeeper.exceptions
import homekeeper.lib


fopen = homekeeper.lib.fopen
lib = homekeeper.lib
ConfigException = homekeeper.exceptions.ConfigException
JSONDecodeError = json.JSONDecodeError


class ConfigData:
    def __init__(self):
        self.directories = []
        self.excludes = [
            '.git',
            '.gitignore',
            'LICENSE',
            'LICENSE.txt',
            'README',
            'README.md',
            'README.txt'
        ]


def get_default_location():
    return os.path.join(lib.get_home_directory(), '.homekeeper.json')


def read(config_file):
    try:
        if not config_file or not os.path.exists(config_file):
            config_file = os.path.abspath(os.path.join(os.getcwd(), '.homekeeper.json'))
            logging.info("trying configuration from current directory: %s", config_file)
        if not os.path.exists(config_file):
            config_file = get_default_location()
            logging.info("trying configuration from home directory :%s", config_file)
        if not os.path.exists(config_file):
            raise ConfigException("could not find configuration file anywhere; create .homekeeper.json first")
        return __read(config_file)
    except IOError as e:
        raise ConfigException("couldn't read from file '{}': {}".format(config_file, e))
    except JSONDecodeError as e:
        raise ConfigException("configuration format in file '{}' was invalid: {}".format(config_file, e))


def write(config_data, config_file=None):
    try:
        if not config_file:
            logging.info("writing configuration to default location: %s", get_default_location())
            config_file = get_default_location()
        return __write(config_data, config_file=config_file)
    except IOError as e:
        raise ConfigException("couldn't write to file '{}': {}".format(config_file, e))


def __read(config_file):
    with fopen(config_file, encoding='utf-8') as f:
        logging.debug("reading configuration from: %s", os.path.abspath(config_file))
        data = json.loads(f.read())
        logging.debug("read configuration: %s", data)
        config_data = ConfigData()
        __read_directories(config_data, data)
        __read_excludes(config_data, data)
        if not config_data.directories:
            raise ConfigException("no configuration key named 'directory' or 'directories' "
                                  "found: {}".format(config_file))
        if not config_data.excludes:
            config_data.excludes = []
        return config_data


def __write(config_data, config_file=None):
    with fopen(config_file, 'w', encoding='utf-8') as f:
        data = dict()
        data['directories'] = config_data.directories
        data['excludes'] = config_data.excludes
        json_data = json.dumps(data, sort_keys=True, indent=4)
        logging.debug("writing configuration: %s", json_data)
        f.write(json_data)
        logging.debug("wrote configuration to: %s", config_file)
        return config_file


def __read_directories(config_data, data):
    if 'directory' in data and 'directories' in data:
        raise ConfigException("configuration key 'directory' cannot be present with key 'directories'")
    if 'directories' in data and isinstance(data['directories'], list):
        config_data.directories = data['directories']
        for directory in config_data.directories:
            if not isinstance(directory, str):
                raise ConfigException("configuration key 'directories' did not contain a string: {}"
                                      .format(directory))
        return
    if 'directories' in data and isinstance(data['directories'], str):
        config_data.directories = [data['directories']]
        return
    if 'directory' in data and isinstance(data['directory'], str):
        config_data.directories = [data['directory']]
        return
    raise ConfigException("configuration key 'directories' and/or 'directory' is not a list or string")


def __read_excludes(config_data, data):
    if 'excludes' not in data:
        config_data.excludes = []
        return
    if not isinstance(data['excludes'], list):
        raise ConfigException("configuration key 'excludes' is not a list: {}".format(data['excludes']))
    config_data.excludes = data['excludes']
    for exclude in config_data.excludes:
        if not isinstance(exclude, str):
            raise ConfigException("configuration key 'excludes' did not contain a string: {}".format(exclude))



import pytest
import homekeeper.lib
import homekeeper.testing


cd = homekeeper.lib.cd
config = homekeeper.config
lib = homekeeper.lib
BaseFakeFilesystemTest = homekeeper.testing.BaseFakeFilesystemTest
ConfigData = homekeeper.config.ConfigData
Homekeeper = homekeeper.Homekeeper


class HomekeeperTest(BaseFakeFilesystemTest):
    @pytest.fixture
    def ctx(self):
        return {
            'cleanup_symlinks': True,
            'config_path': '',
            'overwrite': True,
        }

    # pylint: disable=unused-argument
    def test_simple_usage(self, ctx, os, files, directories):
        config_data = ConfigData()
        config_data.directories = [os.path.join(lib.get_home_directory(), 'dotfiles')]
        config_file = config.write(config_data)
        assert os.path.exists(config_file)
        h = Homekeeper(ctx)
        h.keep()
        h.unkeep()
        h.cleanup()
        assert h.version() == homekeeper.__version__
        for pathname in os.listdir(lib.get_home_directory()):
            assert not os.path.islink(pathname)



import logging
import os
import shutil
import homekeeper.lib


config = homekeeper.config
lib = homekeeper.lib
ConfigData = homekeeper.config.ConfigData


def cleanup_symlinks(*args):
    directory = os.path.join(*args)
    logging.info("removing broken symlinks in directory: %s", directory)
    for item in os.listdir(directory):
        pathname = os.path.join(directory, item)
        basename = os.path.basename(pathname)
        if not os.path.islink(pathname):
            logging.debug("skipping because path is not a symlink: %s", pathname)
            continue
        if os.path.exists(os.readlink(pathname)):
            logging.debug("skipping because path is not a broken symlink: %s", pathname)
            continue
        if not basename.startswith('.'):
            logging.debug("skipping because path does not start with dot: %s", pathname)
            continue
        logging.info("removing broken link: %s", pathname)
        os.unlink(pathname)
    logging.info("finished removing symlinks in directory: %s", directory)


def create_symlink(source, target, overwrite=True):
    dirname = os.path.dirname(target)
    if not os.path.exists(dirname):
        lib.makedirs(dirname)
    if source == target:
        logging.info("skipping because source and target are the same: %s", source)
        return
    if os.path.exists(target) and not overwrite:
        logging.info("skipping because target would be overwritten: %s", target)
        return
    lib.remove(target)
    os.symlink(source, target)
    logging.info('symlinked %s -> %s', source, target)


def create_symlinks(source_directory, target_directory, config_data=ConfigData(), overwrite=True):
    lib.process_directories(
        source_directory,
        target_directory,
        overwrite=overwrite,
        config_data=config_data,
        process=create_symlink)


def restore_symlink(source, target, overwrite=None): # pylint: disable=unused-argument
    if source == target:
        logging.info("skipping because source and target are the same: %s", source)
        return
    if not os.path.islink(target):
        logging.info("skipping because target is not a symlink: %s", target)
        return
    if os.readlink(target) != source:
        logging.info("skipping because target is not symlinked to source: %s", target)
        return
    lib.remove(target)
    if os.path.isfile(source):
        shutil.copyfile(source, target)
        logging.info("restored file %s -> %s", source, target)
    elif os.path.isdir(source):
        shutil.copytree(source, target, symlinks=True)
        logging.info("restored directory %s -> %s", source, target)
    else:
        logging.info("skipping because target is not a file or directory: %s", target)


def restore_symlinks(source_directory, target_directory, config_data=ConfigData()):
    lib.process_directories(
        source_directory,
        target_directory,
        config_data=config_data,
        process=restore_symlink)



import homekeeper.lib
import homekeeper.symlink
import homekeeper.testing

config = homekeeper.config
lib = homekeeper.lib
symlink = homekeeper.symlink
BaseFakeFilesystemTest = homekeeper.testing.BaseFakeFilesystemTest
ConfigData = homekeeper.config.ConfigData


class SymlinkCreateTest(BaseFakeFilesystemTest):
    def test_create_symlink_with_file(self, os):
        source = os.path.join(self.dotfiles_directory, '.bashrc')
        target = os.path.join(self.home_directory, '.bashrc')
        symlink.create_symlink(source, target)
        assert os.path.exists(target)
        assert os.readlink(target) == source

    def test_create_symlink_with_file_overwrites_target(self, os):
        source = os.path.join(self.dotfiles_directory, '.bashrc')
        target = os.path.join(self.home_directory, '.bashrc')
        symlink.create_symlink(source, target)
        source = os.path.join(self.dotfiles_directory, '.bash_profile')
        lib.touch(source)
        symlink.create_symlink(source, target, overwrite=True)
        assert os.readlink(target) == source

    def test_create_symlink__with_file_does_not_overwrite_target(self, os):
        source = os.path.join(self.dotfiles_directory, '.bashrc')
        target = os.path.join(self.home_directory, '.bashrc')
        lib.touch(source)
        symlink.create_symlink(source, target)
        source = os.path.join(self.dotfiles_directory, '.bash_profile')
        lib.touch(source)
        symlink.create_symlink(source, target, overwrite=False)
        assert os.readlink(target) != source

    def test_create_symlink_with_file_that_does_not_exist(self, os):
        source = os.path.join(self.dotfiles_directory, '.bashrc')
        target = os.path.join(self.home_directory, '.bashrc')
        lib.remove(source)
        assert not os.path.exists(source)
        assert not os.path.islink(target)

    def test_create_symlink_with_same_source_and_target(self, os):
        source = os.path.join(self.dotfiles_directory, '.bashrc')
        target = os.path.join(self.dotfiles_directory, '.bashrc')
        symlink.create_symlink(source, target)
        assert os.path.exists(source)
        assert os.path.exists(target)
        assert not os.path.islink(target)

    def test_create_symlink_with_directory(self, os):
        source = os.path.join(self.dotfiles_directory, '.vim')
        target = os.path.join(self.home_directory, '.vim')
        symlink.create_symlink(source, target)
        assert os.path.islink(target)
        assert os.readlink(target) == source

    def test_create_symlink_with_symlink(self, os):
        source = os.path.join(self.dotfiles_directory, '.vimrc.symlink')
        target = os.path.join(self.home_directory, '.vimrc')
        symlink.create_symlink(source, target)
        assert os.path.islink(target)
        assert os.readlink(target) == source

    def test_create_symlinks(self, os, files, directories):
        config_data = ConfigData()
        symlink.create_symlinks(self.dotfiles_directory, self.home_directory, config_data=config_data, overwrite=True)
        for pathname in files + directories:
            source = os.path.join(self.dotfiles_directory, pathname)
            target = os.path.join(self.home_directory, pathname)
            if pathname not in config_data.excludes:
                assert os.path.exists(source)
                assert os.path.exists(target)
                assert os.readlink(target) == source
            elif os.path.exists(source):
                assert not os.path.exists(target)

    def test_create_symlinks_does_not_overwrite(self, os, files, directories):
        for filename in files:
            lib.touch(self.home_directory, filename)
        for dirname in directories:
            lib.makedirs(self.home_directory, dirname)
        config_data = ConfigData()
        symlink.create_symlinks(self.dotfiles_directory, self.home_directory, config_data=config_data, overwrite=False)
        for pathname in files + directories:
            source = os.path.join(self.dotfiles_directory, pathname)
            target = os.path.join(self.home_directory, pathname)
            assert os.path.exists(source)
            assert os.path.exists(target)
            assert not os.path.islink(target)

    def test_create_symlinks_excludes_paths(self, os, files, directories):
        config_data = ConfigData()
        config_data.excludes = files + directories
        symlink.create_symlinks(self.dotfiles_directory, self.home_directory, config_data=config_data, overwrite=True)
        for pathname in files + directories:
            source = os.path.join(self.dotfiles_directory, pathname)
            target = os.path.join(self.home_directory, pathname)
            assert os.path.exists(source)
            assert not os.path.exists(target)



import homekeeper
import setuptools


with open('README.txt', encoding='utf-8') as f:
    long_description = f.read()


setuptools.setup(
    author='Min Huang',
    author_email='min.huang@alumni.usc.edu',
    description='Symlinks your dotfiles from anywhere to your home directory.',
    install_requires=['click >= 7.1.2'],
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    name='homekeeper',
    packages=['homekeeper'],
    platforms=['any'],
    url='https://github.com/retiman/homekeeper',
    version=homekeeper.__version__,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Utilities',
    ],
    entry_points="""
        [console_scripts]
            homekeeper = homekeeper.cli:main
    """,
)
